# encoding: utf-8
# Copyright (C) 2017 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.
'''
pyxtuml is a python library for parsing, manipulating, and generating
BridgePoint xtUML models.
'''
from . import version

from .tools import IdGenerator
from .tools import UUIDGenerator
from .tools import IntegerGenerator
from .tools import OrderedSet

from .tools import Walker
from .tools import Visitor
from .tools import NodePrintVisitor

from .load import load_metamodel
from .load import ParsingException
from .load import ModelLoader

from .persist import persist_database
from .persist import persist_instances
from .persist import persist_schema
from .persist import persist_unique_identifiers

from .persist import serialize_database
from .persist import serialize_schema
from .persist import serialize_class
from .persist import serialize_classes
from .persist import serialize_association
from .persist import serialize_associations
from .persist import serialize_unique_identifiers
from .persist import serialize_instances
from .persist import serialize_instance
from .persist import serialize_value
from .persist import serialize

from .meta import Association
from .meta import Link

from .meta import QuerySet
from .meta import Class
from .meta import BaseObject
from .meta import MetaClass
from .meta import MetaModel

from .meta import MetaException
from .meta import DeleteException
from .meta import RelateException
from .meta import UnrelateException
from .meta import MetaModelException
from .meta import UnknownLinkException
from .meta import UnknownClassException

from .meta import navigate_any
from .meta import navigate_one
from .meta import navigate_many
from .meta import navigate_subtype
from .meta import relate
from .meta import unrelate
from .meta import delete
from .meta import cardinality
from .meta import where_eq
from .meta import sort_reflexive
from .meta import get_metaclass
from .meta import get_metamodel
from .meta import order_by
from .meta import reverse_order_by

from .consistency_check import check_association_integrity
from .consistency_check import check_uniqueness_constraint
from .consistency_check import check_subtype_integrity


#
# Suppress false import warning when invoking python3 -m xtuml.consistency_check.
# For details, see http://bugs.python.org/issue27487
#
import warnings
import sys

if not sys.warnoptions: # Overridable with -W flag
    warnings.filterwarnings('ignore', category=RuntimeWarning, module='runpy')




# encoding: utf-8
# Copyright (C) 2017 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.

import logging

from xtuml import navigate_one as one
from xtuml import navigate_any as any
from xtuml import navigate_many as many
from xtuml import navigate_subtype as subtype

from xtuml.tools import Walker


logger = logging.getLogger(__name__)


class ActionTextGenWalker(Walker):
    '''
    Walk the bridgepoint metamodel and translate action code
    to its textual representation.
    '''
    def __init__(self, level=-1):
        Walker.__init__(self)
        self._buf = ''
        self._lvl = level
        
    def __str__(self):
        return self._buf
    
    def __repr__(self):
        return self._buf
    
    def buf(self, value, *args):
        self._buf += value
        self._buf += ''.join(args)
        
    def buf_linebreak(self, *args):
        self._buf += ''.join(args)
        self._buf += '\n'
        self._buf += '    ' * self._lvl

    def default_accept(self, inst):
        print(inst.__class__.__name__)       
        
    def accept_S_BRG(self, inst):
        self.accept(one(inst).ACT_BRB[697].ACT_ACT[698]())
    
    def accept_O_TFR(self, inst):
        self.accept(one(inst).ACT_OPB[696].ACT_ACT[698]())
    
    def accept_S_SYNC(self, inst):
        self.accept(one(inst).ACT_FNB[695].ACT_ACT[698]())
    
    def accept_O_DBATTR(self, inst):
        self.accept(one(inst).ACT_DAB[693].ACT_ACT[698]())
    
    def accept_SM_ACT(self, inst):
        self.accept(one(inst).ACT_SAB[691].ACT_ACT[698]())
        self.accept(one(inst).ACT_TAB[688].ACT_ACT[698]())
    
    def accept_SPR_PO(self, inst):
        self.accept(one(inst).ACT_POB[687].ACT_ACT[698]())
    
    def accept_SPR_PS(self, inst):
        self.accept(one(inst).ACT_PSB[686].ACT_ACT[698]())
    
    def accept_SPR_RO(self, inst):
        self.accept(one(inst).ACT_ROB[685].ACT_ACT[698]())
    
    def accept_SPR_RS(self, inst):
        self.accept(one(inst).ACT_RSB[684].ACT_ACT[698]())
    
    def accept_ACT_ACT(self, inst):
        self.accept(one(inst).ACT_BLK[666]())
    
    def accept_ACT_BLK(self, inst):
        self._lvl += 1
        if self._lvl:
            self.buf_linebreak()
        
        first_filter = lambda sel: (not one(sel).ACT_SMT[661, 'precedes']() and
                                    not one(sel).ACT_EL[603]() and
                                    not one(sel).ACT_E[603]())
        
        act_smt = one(inst).ACT_SMT[602](first_filter)
        while act_smt:
            self.accept(act_smt)
            act_smt = one(act_smt).ACT_SMT[661, 'succeeds']()
            
        self._lvl -= 1
        self.buf_linebreak()
        
    def accept_ACT_SMT(self, inst):
        self.accept(subtype(inst, 603))
        self.buf_linebreak(';')
        
    def accept_ACT_RET(self, inst):
        self.buf('return ')
        self.accept(one(inst).V_VAL[668]())
        
    def accept_ACT_BRK(self, inst):
        self.buf('break')
        
    def accept_ACT_CON(self, inst):
        self.buf('continue')
        
    def accept_ACT_CTL(self, inst):
        self.buf('control stop')
        
    def accept_ACT_CR(self, inst):
        self.buf('create object instance ')
        self.accept(one(inst).V_VAR[633]())
        
        o_obj = one(inst).O_OBJ[671]()
        self.buf(' of ', o_obj.Key_Lett)
        
    def accept_ACT_CNV(self, inst):
        o_obj = one(inst).O_OBJ[672]()
        self.buf('create object instance of ', o_obj.Key_Lett)
        
    def accept_ACT_DEL(self, inst):
        self.buf('delete object instance ')
        self.accept(one(inst).V_VAR[634]())
        
    def accept_ACT_REL(self, inst):
        self.buf('relate ')
        self.accept(one(inst).V_VAR[615]())
        
        self.buf(' to ')
        self.accept(one(inst).V_VAR[616]())
        
        r_rel = one(inst).R_REL[653]()
        self.buf(' across R', str(r_rel.Numb))
        if inst.relationship_phrase:
            self.buf('.', inst.relationship_phrase)
        
    def accept_ACT_RU(self, inst):
        self.buf('relate ')
        self.accept(one(inst).V_VAR[617]())
        
        self.buf(' to ')
        self.accept(one(inst).V_VAR[618]())
        
        r_rel = one(inst).R_REL[654]()
        self.buf(' across R', str(r_rel.Numb))
        if inst.relationship_phrase:
            self.buf('.', inst.relationship_phrase)
            
        self.buf(' using ')
        self.accept(one(inst).V_VAR[619]())
        
    def accept_ACT_UNR(self, inst):
        self.buf('unrelate ')
        self.accept(one(inst).V_VAR[620]())
        
        self.buf(' from ')
        self.accept(one(inst).V_VAR[621]())
        
        r_rel = one(inst).R_REL[655]()
        self.buf(' across R', str(r_rel.Numb))
        if inst.relationship_phrase:
            self.buf('.', inst.relationship_phrase)
            
    def accept_ACT_URU(self, inst):
        self.buf('unrelate ')
        self.accept(one(inst).V_VAR[622]())
        
        self.buf(' from ')
        self.accept(one(inst).V_VAR[623]())
        
        r_rel = one(inst).R_REL[656]()
        self.buf(' across R', str(r_rel.Numb))
        if inst.relationship_phrase:
            self.buf('.', inst.relationship_phrase)
            
        self.buf(' using ')
        self.accept(one(inst).V_VAR[624]())
        
    def accept_ACT_FIO(self, inst):
        o_obj = one(inst).O_OBJ[677]()
        self.buf('select ', inst.cardinality, ' ')
        self.accept(one(inst).V_VAR[639]())
        self.buf(' from instances of ', o_obj.Key_Lett)
        
    def accept_ACT_SEL(self, inst):
        self.buf('select ', inst.cardinality, ' ')
        self.accept(one(inst).V_VAR[638]())
        self.buf(' related by ')
        self.accept(one(inst).V_VAL[613]())
        self.accept(one(inst).ACT_LNK[637]())
        self.accept(one(inst).ACT_SRW[664]())

    def accept_ACT_SRW(self, inst):
        self.buf('where ')
        self.accept(one(inst).V_VAL[611]())
        
    def accept_ACT_FIW(self, inst):
        o_obj = one(inst).O_OBJ[676]()
        self.buf('select ', inst.cardinality, ' ')
        self.accept(one(inst).V_VAR[665]())
        self.buf(' from instances of ', o_obj.Key_Lett)
        self.buf(' where ')
        self.accept(one(inst).V_VAL[610]())
        
    def accept_ACT_LNK(self, inst):
        o_obj = one(inst).O_OBJ[678]()
        r_rel = one(inst).R_REL[681]()
        self.buf('->', o_obj.Key_Lett, '[R', str(r_rel.Numb))
        if inst.Rel_Phrase:
            self.buf('.', inst.Rel_Phrase)
        self.buf(']')
        self.accept(one(inst).ACT_LNK[604, 'precedes']())
        
    def accept_ACT_AI(self, inst):
        if one(inst).V_VAL[609].V_MSV[801]():
            self.buf('send ')
        else:
            self.buf('assign ')
            
        self.accept(one(inst).V_VAL[689]())
        self.buf(' = ')
        self.accept(one(inst).V_VAL[609]())

    def accept_ACT_WHL(self, inst):
        self.buf('while ')
        self.accept(one(inst).V_VAL[626]())
        self.accept(one(inst).ACT_BLK[608]())
        self.buf('end while')
        
    def accept_ACT_IF(self, inst):
        by_position = lambda inst: (one(inst).ACT_SMT[603]().LineNumber,
                                    one(inst).ACT_SMT[603]().StartPosition)
        
        self.buf('if ')
        self.accept(one(inst).V_VAL[625]())
        self.accept(one(inst).ACT_BLK[607]())
        
        for act_el in sorted(many(inst).ACT_EL[682](), key=by_position):
            self.accept(act_el)
        
        self.accept(one(inst).ACT_E[683]())
        
        self.buf('end if')
        
    def accept_ACT_EL(self, inst):
        self.buf('elif ')
        self.accept(one(inst).V_VAL[659]())
        self.accept(one(inst).ACT_BLK[658]())
        
    def accept_ACT_E(self, inst):
        self.buf('else')
        self.accept(one(inst).ACT_BLK[606]())

    def accept_ACT_FOR(self, inst):
        self.buf('for each ')
        self.accept(one(inst).V_VAR[614]())
        self.buf(' in ')
        self.accept(one(inst).V_VAR[652]())
        self.accept(one(inst).ACT_BLK[605]())
        self.buf('end for')

    def accept_ACT_FNC(self, inst):
        s_sync = one(inst).S_SYNC[675]()
        self.buf('::', s_sync.Name)
        self.buf('(')
        first_filter = lambda sel: one(sel).V_PAR[816, 'succeeds']() is None
        self.accept(any(inst).V_PAR[669](first_filter))
        self.buf(')')
        
    def accept_ACT_BRG(self, inst):
        s_brg = one(inst).S_BRG[674]()
        s_ee = one(s_brg).S_EE[19]()
        self.buf('bridge ', s_ee.Key_Lett, '::', s_brg.Name)
        self.buf('(')
        first_filter = lambda sel: one(sel).V_PAR[816, 'succeeds']() is None
        self.accept(any(inst).V_PAR[628](first_filter))
        self.buf(')')
        
    def accept_ACT_IOP(self, inst):
        self.accept(one(inst).SPR_PO[680].SPR_PEP[4503]())
        self.accept(one(inst).SPR_RO[657].SPR_REP[4502]())
        self.buf('(')
        first_filter = lambda sel: one(sel).V_PAR[816, 'succeeds']() is None
        self.accept(any(inst).V_PAR[679](first_filter))
        self.buf(')')
        
    def accept_ACT_SGN(self, inst):
        self.accept(one(inst).SPR_PS[663].SPR_PEP[4503]())
        self.accept(one(inst).SPR_RS[660].SPR_REP[4502]())
        self.buf('(')
        first_filter = lambda sel: one(sel).V_PAR[816, 'succeeds']() is None
        self.accept(any(inst).V_PAR[662](first_filter))
        self.buf(')')
        
    def accept_ACT_TFM(self, inst):
        o_tfr = one(inst).O_TFR[673]()
        o_obj = one(o_tfr).O_OBJ[115]()
        v_var = one(inst).V_VAR[667]()
        
        if not v_var:
            self.buf(o_obj.Key_Lett, "::", o_tfr.Name)
        else:
            self.buf('transform ')
            self.accept(v_var)
            self.buf('.', o_tfr.Name)
            
        self.buf('(')
        first_filter = lambda sel: one(sel).V_PAR[816, 'succeeds']() is None
        self.accept(any(inst).V_PAR[627](first_filter))
        self.buf(')')
        
    def accept_V_VAL(self, inst):
        self.accept(subtype(inst, 801))
            
    def accept_V_TVL(self, inst):
        self.accept(one(inst).V_VAR[805]())

    def accept_V_ISR(self, inst):
        self.accept(one(inst).V_VAR[809]())
        
    def accept_V_VAR(self, inst):
        self.buf(inst.Name)
        
    def accept_V_IRF(self, inst):
        self.accept(one(inst).V_VAR[808]())
        
    def accept_V_PVL(self, inst):
        self.buf('param.')
        self.accept(one(inst).S_BPARM[831]())
        self.accept(one(inst).S_SPARM[832]())
        self.accept(one(inst).O_TPARM[833]())
        self.accept(one(inst).C_PP[843]())

    def accept_V_SLR(self, inst):
        self.buf('selected')
        
    def accept_V_MVL(self, inst):
        self.accept(one(inst).V_VAL[837]())
        s_mbr = one(inst).S_MBR[836]()
        self.buf('.', s_mbr.Name)
        
    def accept_V_AVL(self, inst):
        self.accept(one(inst).V_VAL[807]())
        o_attr = one(inst).O_ATTR[806]()
        self.buf('.', o_attr.Name)
        
    def accept_V_AER(self, inst):
        self.accept(one(inst).V_VAL[838]())
        self.buf('[')
        self.accept(one(inst).V_VAL[839]())
        self.buf(']')
        
    def accept_V_LIN(self, inst):
        self.buf(inst.Value)
        
    def accept_V_LRL(self, inst):
        self.buf(inst.Value)
        
    def accept_V_LST(self, inst):
        self.buf('"%s"' % inst.Value)
        
    def accept_V_LBO(self, inst):
        self.buf(inst.Value.lower())

    def accept_V_LEN(self, inst):
        s_enum = one(inst).S_ENUM[824]()
        s_dt = one(s_enum).S_EDT[27].S_DT[17]()
        self.buf(s_dt.Name, '::', s_enum.Name)
        
    def accept_V_BIN(self, inst):
        self.buf('(')
        self.accept(one(inst).V_VAL[802]())
        self.buf(' ', inst.Operator, ' ')
        self.accept(one(inst).V_VAL[803]())
        self.buf(')')

    def accept_V_UNY(self, inst):
        self.buf('(')
        self.buf(inst.Operator, ' ')
        self.accept(one(inst).V_VAL[804]())
        self.buf(')')

    def accept_V_PAR(self, inst):
        if one(inst).V_PAR[816, 'succeeds']() is not None:
            self.buf(', ')
            
        self.buf(inst.Name, ': ')
        self.accept(one(inst).V_VAL[800]())
        self.accept(one(inst).V_PAR[816, 'precedes']())

    def accept_V_EDV(self, inst):
        self.buf('rcvd_evt.')
        self.accept(any(inst).V_EPR[834]())  # TODO: this is a many-relation, why?
        
    def accept_V_EPR(self, inst):
        self.accept(one(inst).SM_EVTDI[846]())
        self.accept(one(inst).C_PP[847]())
        
    def accept_SM_EVTDI(self, inst):
        self.buf(inst.Name)
    
    def accept_V_FNV(self, inst):
        s_sync = one(inst).S_SYNC[827]()
        self.buf('::', s_sync.Name)
        self.buf('(')
        first_filter = lambda sel: one(sel).V_PAR[816, 'succeeds']() is None
        self.accept(any(inst).V_PAR[817](first_filter))
        self.buf(')')
        
    def accept_V_BRV(self, inst):
        s_brg = one(inst).S_BRG[828]()
        s_ee = one(s_brg).S_EE[19]()
        self.buf('bridge ', s_ee.Key_Lett, '::', s_brg.Name)
        self.buf('(')
        first_filter = lambda sel: one(sel).V_PAR[816, 'succeeds']() is None
        self.accept(any(inst).V_PAR[810](first_filter))
        self.buf(')')
        
    def accept_V_TRV(self, inst):
        o_tfr = one(inst).O_TFR[829]()
        o_obj = one(o_tfr).O_OBJ[115]()
        v_var = one(inst).V_VAR[830]()
        
        if not v_var:
            self.buf(o_obj.Key_Lett, "::", o_tfr.Name)
        else:
            self.buf('transform ')
            self.accept(v_var)
            self.buf('.', o_tfr.Name)
            
        self.buf('(')
        first_filter = lambda sel: one(sel).V_PAR[816, 'succeeds']() is None
        self.accept(any(inst).V_PAR[811](first_filter))
        self.buf(')')
        
    def accept_V_MSV(self, inst):
        self.accept(one(inst).SPR_PEP[841]())
        self.accept(one(inst).SPR_REP[845]())
        self.buf('(')
        first_filter = lambda sel: one(sel).V_PAR[816, 'succeeds']() is None
        self.accept(any(inst).V_PAR[842](first_filter))
        self.buf(')')

    def accept_V_SCV(self, inst):
        cnst_syc = one(inst).CNST_SYC[850]()
        cnst_csp = one(cnst_syc).CNST_CSP[1504]()
        self.buf(cnst_csp.InformalGroupName, '::', cnst_syc.Name)
        
    def accept_SPR_PEP(self, inst):
        c_po = one(inst).C_P[4501].C_IR[4009].C_PO[4016]()
        c_ep = one(inst).C_EP[4501]()
        self.buf(c_po.Name, '::', c_ep.Name)
        
    def accept_SPR_REP(self, inst):
        c_po = one(inst).C_R[4500].C_IR[4009].C_PO[4016]()
        c_ep = one(inst).C_EP[4500]()
        self.buf(c_po.Name, '::', c_ep.Name)
        
    def accept_E_GPR(self, inst):
        self.buf('generate ')
        self.accept(one(inst).V_VAL[714]())
    
    def accept_E_ESS(self, inst):
        self.accept(one(inst).E_CES[701]())
        self.accept(one(inst).E_GES[701]())

        self.buf('(')
        first_filter = lambda sel: one(sel).V_PAR[816, 'succeeds']() is None
        self.accept(any(inst).V_PAR[700](first_filter))
        self.buf(')')
        
        self.accept(one(inst).E_CES[701].E_CSME[702]())
        self.accept(one(inst).E_CES[701].E_CEE[702]())
        
        self.accept(one(inst).E_GES[701].E_GSME[703]())
        self.accept(one(inst).E_GES[701].E_GEE[703]())
        
    def accept_E_GES(self, inst):
        self.buf('generate ')
        self.accept(one(inst).E_GSME[703].SM_EVT[707]())
        self.accept(one(inst).E_GEE[703].S_EEEVT[709]())
        
    def accept_E_CES(self, inst):
        self.buf('create event instance ')
        self.accept(one(inst).V_VAR[710]())
        
        self.buf(' of ')
        
        self.accept(one(inst).E_CSME[702].SM_EVT[706]())
        self.accept(one(inst).E_CEE[702].S_EEEVT[708]())
                
    def accept_SM_EVT(self, inst):
        self.buf(inst.Drv_Lbl)
        if one(inst).SM_PEVT[525]():
            self.buf('*')
            
        self.buf(":'%s'" % inst.Mning)
        
    def accept_E_GSME(self, inst):
        self.buf(' to ')
        self.accept(one(inst).E_GEN[705].V_VAR[712]())
        self.accept(one(inst).E_GAR[705]())
        self.accept(one(inst).E_GEC[705]())
        
    def accept_E_GAR(self, inst):
        o_obj = one(inst).E_GSME[705].SM_EVT[707].SM_SM[502].SM_ASM[517].O_OBJ[519]()
        self.buf(o_obj.Key_Lett, ' class')
        
    def accept_E_GEC(self, inst):
        o_obj = one(inst).E_GSME[705].SM_EVT[707].SM_SM[502].SM_ISM[517].O_OBJ[518]()
        self.buf(o_obj.Key_Lett, ' creator')
        
    def accept_E_CSME(self, inst):
        self.buf(' to ')
        self.accept(one(inst).E_CEI[704].V_VAR[711]())
        self.accept(one(inst).E_CEA[704]())
        self.accept(one(inst).E_CEC[704]())
        
    def accept_E_CEA(self, inst):
        o_obj = one(inst).E_CSME[704].SM_EVT[706].SM_SM[502].SM_ASM[517].O_OBJ[519]()
        self.buf(o_obj.Key_Lett, ' class')
        
    def accept_E_CEC(self, inst):
        o_obj = one(inst).E_CSME[704].SM_EVT[706].SM_SM[502].SM_ISM[517].O_OBJ[518]()
        self.buf(o_obj.Key_Lett, ' creator')
        
    def accept_O_TPARM(self, inst):
        self.buf(inst.Name)

    def accept_S_SPARM(self, inst):
        self.buf(inst.Name)
        
    def accept_S_BPARM(self, inst):
        self.buf(inst.Name)
        
    def accept_C_PP(self, inst):
        self.buf(inst.Name)
    

def gen_text_action(instance):
    '''
    Generate textual OAL action code from an *instance* in the BridgePoint 
    metamodel. The input may be an instance of the following classes:

    - S_SYNC
    - S_BRG
    - O_TFR
    - O_DBATTR
    - SM_ACT
    - SPR_RO
    - SPR_RS
    - SPR_PO
    - SPR_PS
        
    In addition, anything in the ooaofooa subsystems Value or Body, e.g. ACT_SMT
    or V_VAL are also supported.
    '''
    w = ActionTextGenWalker(-1)
    w.accept(instance)
    return str(w)




# encoding: utf-8
# Copyright (C) 2017 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.

'''
Loading support for xtUML models (based on sql).
'''

import uuid
import logging
import os
import re

from ply import lex
from ply import yacc

import xtuml


logger = logging.getLogger(__name__)


def guess_type_name(value):
    '''
    Guess the type name of a serialized value.
    '''
    value = str(value)
    
    if value.upper() in ['TRUE', 'FALSE']:
        return 'BOOLEAN'
    
    elif re.match(r'(-)?(\d+)(\.\d+)', value):
        return 'REAL'
    
    elif re.match(r'(-)?(\d+)', value):
        return 'INTEGER'
    
    elif re.match(r'\'((\'\')|[^\'])*\'', value):
        return 'STRING'
    
    elif re.match(r'\"([^\\\n]|(\\.))*?\"', value):
        return 'UNIQUE_ID'


def deserialize_value(ty, value):
    '''
    Deserialize a value of some type
    '''
    uty = ty.upper()
    
    if uty == 'BOOLEAN':
        if value.isdigit():
            return bool(int(value))
        elif value.upper() == 'FALSE':
            return False
        elif value.upper() == 'TRUE':
            return True
        else:
            return None
    
    elif uty == 'INTEGER': 
        if '"' in value:
            return uuid.UUID(value[1:-1]).int
        else:
            return int(value)
    
    elif uty == 'REAL': 
        return float(value)
    
    elif uty == 'STRING': 
        return value[1:-1].replace("''", "'")
    
    elif uty == 'UNIQUE_ID': 
        if '"' in value:
            return uuid.UUID(value[1:-1]).int
        else:
            return int(value)

    
class ParsingException(Exception):
    '''
    An exception that may be thrown while loading (and parsing) a metamodel.
    '''
    pass


class Stmt(object):
    offset = None
    lineno = None
    filename = None


class CreateInstanceStmt(Stmt):
    
    def __init__(self, kind, values, names):
        self.kind = kind
        self.values = values
        self.names = names


class CreateClassStmt(Stmt):
    
    def __init__(self, kind, attributes):
        self.kind = kind
        self.attributes = attributes
        

class CreateAssociationStmt(Stmt):
    def __init__(self, rel_id, source_kind, source_cardinality, source_keys, 
                 source_phrase, target_kind, target_cardinality, target_keys,
                target_phrase):
        self.rel_id = rel_id
        self.source_kind = source_kind
        self.target_kind = target_kind
        self.source_keys = source_keys
        self.target_keys = target_keys
        self.source_cardinality = source_cardinality
        self.target_cardinality = target_cardinality
        self.source_phrase = source_phrase
        self.target_phrase = target_phrase
        
 
class CreateUniqueStmt(Stmt):
    
    def __init__(self, kind, name, attributes):
        self.kind = kind
        self.name = name
        self.attributes = attributes


class ModelLoader(object):
    '''
    Class for loading metamodels previously persisted to disk.
    
    Data may be provided in any order, e.g. instances followed by associations, 
    followed by class definitions. One single loader may be used to build
    several *xtuml.MetaModel* objects, and additional data may be provided at
    any time.
    
    **Note:** Additional data will not affect previosly built metamodels.
    
    Usage example:
    
    >>> l = xtuml.ModelLoader()
    >>> l.filename_input('data.sql')
    >>> l.filename_input('schema.sql')
    >>> m1 = l.build_metamodel()
    >>> l.filename_input('additional_data.sql')
    >>> m2 = l.build_metamodel()
    '''
    reserved = (
        'CREATE',
        'FALSE',
        'FROM',
        'INDEX',
        'INSERT',
        'INTO',
        'ON',
        'PHRASE',
        'REF_ID',
        'ROP',
        'TABLE',
        'TO',
        'TRUE',
        'UNIQUE',
        'VALUES'
    )

    tokens = reserved + (
        'CARDINALITY',
        'COMMA',
        'FRACTION',
        'GUID',
        'ID',
        'LPAREN',
        'MINUS',
        'NUMBER',
        'RPAREN',
        'RELID',
        'SEMICOLON',
        'STRING'
    )

    # A string containing ignored characters (spaces and tabs).
    t_ignore = ' \t\r\x0c'

    parser = None
    lexer = None
    statements = None
    
    def __init__(self):
        self.statements = list()
        self.parser = yacc.yacc(debuglog=logger,
                                errorlog=logger,
                                optimize=1,
                                module=self,
                                outputdir=os.path.dirname(__file__),
                                tabmodule='xtuml.__xtuml_parsetab')
    
    def input(self, data, name='<string>'):
        '''
        Parse *data* directly from a string. The *name* is used when reporting
        positional information if the parser encounter syntax errors.
        '''
        lexer = lex.lex(debuglog=logger,
                        errorlog=logger,
                        optimize=1,
                        module=self,
                        outputdir=os.path.dirname(__file__),
                        lextab="xtuml.__xtuml_lextab")
        lexer.filename = name
        logger.debug('parsing %s' % name)
        s = self.parser.parse(lexer=lexer, input=data, tracking=1)
        self.statements.extend(s)

    def filename_input(self, filename):
        '''
        Open and read from a *filename* on disk, and parse its content.
        '''
        with open(filename, 'r') as f:
            return self.file_input(f)
    
    def file_input(self, file_object):
        '''
        Read and parse data from a *file object*, i.e. the type of object 
        returned by the builtin python function *open()*.
        '''
        return self.input(file_object.read(), name=file_object.name)

    def populate_classes(self, metamodel):
        '''
        Populate a *metamodel* with classes previously encountered from input.
        '''
        for stmt in self.statements:
            if isinstance(stmt, CreateClassStmt):
                metamodel.define_class(stmt.kind, stmt.attributes)

    def populate_associations(self, metamodel):
        '''
        Populate a *metamodel* with associations previously encountered from
        input.
        '''
        for stmt in self.statements:
            if not isinstance(stmt, CreateAssociationStmt):
                continue
            
            ass = metamodel.define_association(stmt.rel_id,
                                         stmt.source_kind,
                                         stmt.source_keys,
                                         'M' in stmt.source_cardinality,
                                         'C' in stmt.source_cardinality,
                                         stmt.source_phrase,
                                         stmt.target_kind,
                                         stmt.target_keys,
                                         'M' in stmt.target_cardinality,
                                         'C' in stmt.target_cardinality,
                                         stmt.target_phrase)
            ass.formalize()

    def populate_unique_identifiers(self, metamodel):
        '''
        Populate a *metamodel* with class unique identifiers previously
        encountered from input.
        '''
        for stmt in self.statements:
            if isinstance(stmt, CreateUniqueStmt):
                metamodel.define_unique_identifier(stmt.kind, stmt.name, 
                                                   *stmt.attributes)

    @staticmethod
    def _populate_matching_class(metamodel, kind, names, values):
        '''
        Populate a *metamodel* with a class that matches the given *insert
        statement*.
        '''
        attributes = list()
        for name, value in zip(names, values):
            ty = guess_type_name(value)
            attr = (name, ty)
            attributes.append(attr)
                
        return metamodel.define_class(kind, attributes)
    
    @staticmethod
    def _populate_instance_with_positional_arguments(metamodel, stmt):
        '''
        Populate a *metamodel* with an instance previously encountered from 
        input that was defined using positional arguments.
        '''
        if stmt.kind.upper() not in metamodel.metaclasses:
            names = ['_%s' % idx for idx in range(len(stmt.values))]
            ModelLoader._populate_matching_class(metamodel, stmt.kind, 
                                                 names, stmt.values)
            
        metaclass = metamodel.find_metaclass(stmt.kind)
        if len(metaclass.attributes) != len(stmt.values):
            logger.warning('%s:%d:schema mismatch' % (stmt.filename, stmt.lineno))
                
        inst = metamodel.new(stmt.kind)
        for attr, value in zip(metaclass.attributes, stmt.values):
            name, ty = attr
            py_value = deserialize_value(ty, value)
            if py_value is None:
                raise ParsingException("%s:%d:unable to deserialize "\
                                       "%s to a %s" % (stmt.filename,
                                                       stmt.lineno,
                                                       value,
                                                       ty))

            inst.__dict__[name] = py_value
        
        return inst
    
    @staticmethod
    def _populate_instance_with_named_arguments(metamodel, stmt):
        '''
        Populate a *metamodel* with an instance previously encountered from 
        input that was defined using named arguments.
        '''
        if stmt.kind.upper() not in metamodel.metaclasses:
            ModelLoader._populate_matching_class(metamodel, stmt.kind, 
                                                 stmt.names, stmt.values)
            
        metaclass = metamodel.find_metaclass(stmt.kind)
            
        schema_unames = [name.upper() for name in metaclass.attribute_names]
        inst_unames = [name.upper() for name in stmt.names]
        
        if set(inst_unames) - set(schema_unames):
            logger.warning('%s:%d:schema mismatch' % (stmt.filename, stmt.lineno))
            
        inst = metamodel.new(stmt.kind)
        for name, ty in metaclass.attributes:
            uname = name.upper()
            if uname in inst_unames:
                idx = inst_unames.index(uname)
                value = deserialize_value(ty, stmt.values[idx])
                if value is None:
                    raise ParsingException("%s:%d:unable to deserialize "\
                                           "%s to a %s" % (stmt.filename,
                                                           stmt.lineno,
                                                           value,
                                                           ty))
            else:
                value = None
            
            inst.__dict__[name] = value

        return inst
    
    def populate_instances(self, metamodel):
        '''
        Populate a *metamodel* with instances previously encountered from
        input.
        '''
        for stmt in self.statements:
            if not isinstance(stmt, CreateInstanceStmt):
                continue
            
            if stmt.names:
                fn = self._populate_instance_with_named_arguments
            else:
                fn = self._populate_instance_with_positional_arguments
            
            fn(metamodel, stmt)
    
    def populate_connections(self, metamodel):
        '''
        Populate links in a *metamodel* with connections between them.
        '''
        storage = dict()
        for ass in metamodel.associations:
            source_class = ass.source_link.to_metaclass
            target_class = ass.target_link.to_metaclass

            if target_class not in storage:
                storage[target_class] = dict()
            
            link_key = frozenset(ass.source_link.key_map.values())
            if link_key not in storage[target_class]:
                storage[target_class][link_key] = dict()
                for other_inst in target_class.storage:
                    inst_key = ass.source_link.compute_index_key(other_inst)
                    if inst_key is None:
                        continue
                    
                    if inst_key not in storage[target_class][link_key]:
                        storage[target_class][link_key][inst_key] = xtuml.OrderedSet()

                    storage[target_class][link_key][inst_key].add(other_inst)

            for inst in source_class.storage:
                inst_key = ass.source_link.compute_lookup_key(inst)
                if inst_key is None:
                    continue
                
                if inst_key not in storage[target_class][link_key]:
                    continue
                
                for other_inst in storage[target_class][link_key][inst_key]:
                    ass.source_link.connect(other_inst, inst, check=False)
                    ass.target_link.connect(inst, other_inst, check=False)

        for inst in metamodel.instances:
            metaclass = xtuml.get_metaclass(inst)
            for attr in metaclass.referential_attributes:
                if attr in inst.__dict__:
                    delattr(inst, attr)

    def populate(self, metamodel):
        '''
        Populate a *metamodel* with entities previously encountered from input.
        '''
        self.populate_classes(metamodel)
        self.populate_unique_identifiers(metamodel)
        self.populate_associations(metamodel)
        self.populate_instances(metamodel)
        self.populate_connections(metamodel)

    def build_metamodel(self, id_generator=None):
        '''
        Build and return a *xtuml.MetaModel* containing previously loaded input.
        '''
        m = xtuml.MetaModel(id_generator)
        
        self.populate(m)
        
        return m

    def t_comment(self, t):
        r'\-\-([^\n]*\n?)'
        t.lexer.lineno += (t.value.count("\n"))
        t.endlexpos = t.lexpos + len(t.value)

    def t_COMMA(self, t):
        r','
        t.endlexpos = t.lexpos + len(t.value)
        return t

    def t_FRACTION(self, t):
        r'(\d+)(\.\d+)'
        t.endlexpos = t.lexpos + len(t.value)
        return t

    def t_RELID(self, t):
        r'R[0-9]+'
        t.endlexpos = t.lexpos + len(t.value)
        return t

    def t_CARDINALITY(self, t):
        r'(1C)'
        t.endlexpos = t.lexpos + len(t.value)
        return t
    
    def t_ID(self, t):
        r'[A-Za-z_][\w_]*'
        vup = t.value.upper()
        if vup in self.reserved:
            t.type = vup
        t.endlexpos = t.lexpos + len(t.value)
        return t

    def t_LPAREN(self, t):
        r'\('
        t.endlexpos = t.lexpos + len(t.value)
        return t

    def t_MINUS(self, t):
        r'-'
        t.endlexpos = t.lexpos + len(t.value)
        return t

    def t_NUMBER(self, t):
        r'[0-9]+'
        t.endlexpos = t.lexpos + len(t.value)
        return t

    def t_RPAREN(self, t):
        r'\)'
        t.endlexpos = t.lexpos + len(t.value)
        return t

    def t_SEMICOLON(self, t):
        r';'
        t.endlexpos = t.lexpos + len(t.value)
        return t

    def t_STRING(self, t):
        r'\'((\'\')|[^\'])*\''
        t.lexer.lineno += (t.value.count("\n"))
        t.endlexpos = t.lexpos + len(t.value)
        return t

    def t_GUID(self, t):
        r'\"([^\\\n]|(\\.))*?\"'
        t.endlexpos = t.lexpos + len(t.value)
        return t

    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += len(t.value)
        t.endlexpos = t.lexpos + len(t.value)

    def t_error(self, t):
        raise ParsingException("illegal character '%s' at %s:%d" % (t.value[0],
                               t.lexer.filename, t.lineno))

    def p_empty_translation_unit(self, p):
        '''translation_unit :'''
        p[0] = list()
        
    def p_translation_unit(self, p):
        '''translation_unit : statement_sequence'''
        p[0] = p[1]

    def p_statement_sequence(self, p):
        '''statement_sequence : statement_sequence statement'''
        p[0] = p[1]
        p[0].append(p[2])

    def p_statement_sequence_endpoint(self, p):
        '''statement_sequence : statement'''
        p[0] = [p[1]]
        
    def p_statement(self, p):
        '''
        statement : create_table_statement SEMICOLON
                  | insert_into_statement SEMICOLON
                  | create_rop_statement SEMICOLON
                  | create_index_statement SEMICOLON
        '''
        p[0] = p[1]
        p[0].offset = p.lexpos(1)
        p[0].lineno = p.lineno(1)
        p[0].filename = p.lexer.filename

    def p_create_table_statement(self, p):
        '''create_table_statement : CREATE TABLE identifier LPAREN attribute_sequence RPAREN'''
        p[0] = CreateClassStmt(p[3], p[5])

    def p_empty_attribute_sequence(self, p):
        '''attribute_sequence : '''
        p[0] = []
        
    def p_attribute_sequence_endpoint(self, p):
        '''attribute_sequence : attribute'''
        p[0] = [p[1]]

    def p_attribute_sequence(self, p):
        '''attribute_sequence : attribute_sequence COMMA attribute'''
        p[0] = p[1]
        p[0].append(p[3])

    def p_attribute(self, p):
        '''attribute : identifier identifier'''
        p[0] = (p[1], p[2])

    def p_ordered_insert_into_statement(self, p):
        '''
        insert_into_statement : INSERT INTO identifier VALUES LPAREN value_sequence RPAREN
        '''
        p[0] = CreateInstanceStmt(p[3], p[6], None)

    def p_named_insert_into_statement(self, p):
        '''
        insert_into_statement : INSERT INTO identifier LPAREN identifier_sequence RPAREN VALUES LPAREN value_sequence RPAREN
        '''
        p[0] = CreateInstanceStmt(p[3], p[9], p[5])
    
    def p_empty_value_sequence(self, p):
        '''value_sequence : '''
        p[0] = []

    def p_value_sequence(self, p):
        '''value_sequence : value_sequence COMMA value'''
        p[0] = p[1]
        p[0].append(p[3])
        
    def p_value_sequence_endpoint(self, p):
        '''value_sequence : value'''
        p[0] = [p[1]]
        
    def p_value(self, p):
        '''
        value : FRACTION
              | NUMBER
              | STRING
              | GUID
              | TRUE
              | FALSE
        '''
        p[0] = p[1]

    def p_negative_value(self, p):
        '''
        value : MINUS FRACTION
              | MINUS NUMBER
        '''
        p[0] = p[1] + p[2]
        
    def p_create_rop_statement(self, p):
        '''create_rop_statement : CREATE ROP REF_ID RELID FROM association_end TO association_end'''
        args = [p[4]]
        args.extend(p[6])
        args.extend(p[8])
        p[0] = CreateAssociationStmt(*args)

    def p_association_end(self, p):
        '''association_end : cardinality identifier LPAREN identifier_sequence RPAREN'''
        p[0] = (p[2], p[1], p[4], '')

    def p_phrased_association_end(self, p):
        '''association_end : cardinality identifier LPAREN identifier_sequence RPAREN PHRASE STRING'''
        p[0] = (p[2], p[1], p[4], p[7][1:-1])

    def p_cardinality_1(self, p):
        '''cardinality : NUMBER'''
        if p[1] != '1':
            raise ParsingException("illegal cardinality (%s) at %s:%d" % (p[1],
                                   p.lexer.filename, p.lineno(1)))
        p[0] = p[1]

    def p_cardinality_many(self, p):
        '''cardinality : ID'''
        if p[1] not in ['M', 'MC']:
            raise ParsingException("illegal cardinality (%s) at %s:%d" % (p[1],
                                   p.lexer.filename, p.lineno(1)))
        p[0] = p[1]

    def p_cardinality_01(self, p):
        '''cardinality : CARDINALITY'''
        p[0] = p[1]

    def p_empty_identifier_sequence(self, p):
        '''identifier_sequence : '''
        p[0] = []
        
    def p_identifier_sequence(self, p):
        '''identifier_sequence : identifier_sequence COMMA identifier'''
        p[0] = p[1]
        p[0].append(p[3])

    def p_identifier_sequence_endpoint(self, p):
        '''identifier_sequence : identifier'''
        p[0] = [p[1]]

    def p_identifier(self, p):
        '''
        identifier : ID
                   | CREATE
                   | INSERT
                   | INTO
                   | VALUES
                   | TABLE
                   | ROP
                   | REF_ID
                   | FROM
                   | TO
                   | PHRASE
                   | UNIQUE
                   | INDEX
                   | ON
                   | TRUE
                   | FALSE
        '''
        p[0] = p[1]
        
    def p_create_index_statement(self, p):
        '''
        create_index_statement : CREATE UNIQUE INDEX identifier ON identifier LPAREN identifier_sequence RPAREN
        '''
        p[0] = CreateUniqueStmt(p[6], p[4], p[8])

    def p_error(self, p):
        if p:
            raise ParsingException("illegal token %s (%s) at %s:%d" % (p.type, 
                                   p.value, p.lexer.filename, p.lineno))
        else:
            raise ParsingException("unknown error")


def load_metamodel(resource):
    '''
    Load and return a metamodel from a *resource*. The *resource* may be either
    a filename, or a list of filenames.
    
    Usage example:
    
    >>> metamodel = xtuml.load_metamodel(['schema.sql', 'data.sql'])
    '''
    if isinstance(resource, str):
        resource = [resource]
        
    loader = ModelLoader()
    for filename in resource:
        loader.filename_input(filename)
    
    return loader.build_metamodel()




# encoding: utf-8
# Copyright (C) 2017 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.

name = 'pyxtuml'
date = '2021-11-06'
version = '2.3'
release = '2.3.1a1'  # ensure that this is the same as in setup.py

complete_string = '%s v%s (%s)' % (name, release, date)



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2017 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.
import logging
import unittest
import sys

try:
    from setuptools import setup
    from setuptools import Command
    from setuptools.command.build_py import build_py
except ImportError:
    from distutils.core import setup
    from distutils.core  import Command
    from distutils.command.build_py import build_py


logging.basicConfig(level=logging.DEBUG)


class BuildCommand(build_py):
    
    def run(self):
        import xtuml
        from bridgepoint import oal

        l = xtuml.ModelLoader()
        l.input('', name='<empty string>')
        l.build_metamodel()
        oal.parse('')
        build_py.run(self)


class TestCommand(Command):
    description = "Execute unit tests"
    user_options = [('name=', None, 'Limit testing to a single test case or test method')]

    def initialize_options(self):
        self.name = None
    
    def finalize_options(self):
        if self.name and not self.name.startswith('tests.'):
            self.name = 'tests.' + self.name

    def run(self):
        if self.name:
            suite = unittest.TestLoader().loadTestsFromName(self.name)
        else:
            suite = unittest.TestLoader().discover('tests')
        
        runner = unittest.TextTestRunner(verbosity=2, buffer=True)
        exit_code = not runner.run(suite).wasSuccessful()
        sys.exit(exit_code)


from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.rst')) as f:
    long_description = f.read()

setup(name='pyxtuml',
      version='2.3.1a1', # ensure that this is the same as in xtuml.version
      description='Library for parsing, manipulating, and generating BridgePoint xtUML models',
      long_description=long_description,
      long_description_content_type='text/x-rst',
      author=u'John Törnblom',
      author_email='john.tornblom@gmail.com',
      url='https://github.com/xtuml/pyxtuml',
      license='LGPLv3+',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Topic :: Software Development :: Code Generators',
          'Topic :: Software Development :: Compilers',
          'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6'],
      keywords='xtuml bridgepoint',
      packages=['xtuml', 'bridgepoint'],
      requires=['ply'],
      install_requires=['ply'],
      setup_requires=['ply'],
      cmdclass={'build_py': BuildCommand,
                'test': TestCommand}
      )




# encoding: utf-8
# Copyright (C) 2017 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.

import datetime
import sys
import time


class LOG(object):
    
    @staticmethod
    def LogInteger(message):
        print('LogInteger: %d' % message)

    @staticmethod
    def LogReal(message, r):
        print('LogReal: %f %s' % (r, message))
        
    @staticmethod
    def LogInfo(message):
        print('LogInfo: %s' % message)
        
    @staticmethod
    def LogFailure(message):
        print('LogFailure: %s' % message)
        
    @staticmethod
    def LogSuccess(message):
        print('LogSuccess: %s' % message)
        

class TIM(object):
    
    @staticmethod
    def create_date(day, hour, minute, month, second, year):
        return datetime.datetime(year, month, day, hour, minute, second)
    
    @staticmethod
    def current_date():
        return datetime.datetime.now()
    
    @staticmethod
    def current_clock():
        return time.time()
    
    @staticmethod
    def get_year(date):
        return date.year
    
    @staticmethod
    def get_month(date):
        return date.month
    
    @staticmethod
    def get_day(date):
        return date.day
    
    @staticmethod
    def get_hour(date):
        return date.hour
    
    @staticmethod
    def get_minute(date):
        return date.minute
    
    @staticmethod
    def get_second(date):
        return date.second
    
    
class ARCH(object):
    
    @staticmethod
    def shutdown():
        sys.exit()
        

class PERSIST(object):
    
    @staticmethod
    def commit():
        return 0
    
    @staticmethod
    def restore():
        return 0
    

class NVS(object):
    
    @staticmethod
    def format():
        return 0
    
    @staticmethod
    def version(first, second):
        return 0
    
    @staticmethod
    def checksum(first, second):
        return 0
        



#!/usr/bin/env python
# encoding: utf-8
# Copyright (C) 2017 John Törnblom
#
# This file is part of pyxtuml.
#
# pyxtuml is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# pyxtuml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with pyxtuml. If not, see <http://www.gnu.org/licenses/>.
'''
Generate an xsd schema file for an xtUML model. 
The arguments are either xtuml files, or folders containing *.xtuml files.
Note that some type of attributes are not supported, e.g. instance handles or 
timers.
'''

import sys
import optparse
import logging
import xtuml

import xml.etree.ElementTree as ET
import xml.dom.minidom

from xtuml import navigate_any as nav_any
from xtuml import navigate_any as nav_one
from xtuml import navigate_many as nav_many
from bridgepoint import ooaofooa


logger = logging.getLogger('gen_xsd_schema')


def get_type_name(s_dt):
    '''
    get the xsd name of a S_DT
    '''
    s_cdt = nav_one(s_dt).S_CDT[17]()
    if s_cdt and s_cdt.Core_Typ in range(1, 6):
        return s_dt.Name
    
    s_edt = nav_one(s_dt).S_EDT[17]()
    if s_edt:
        return s_dt.Name
    
    s_udt = nav_one(s_dt).S_UDT[17]()
    if s_udt:
        return s_dt.Name

#    s_sdt = nav_one(s_dt).S_SDT[17]()
#    if s_sdt:
#        return s_dt.name
    
    
def get_refered_attribute(o_attr):
    '''
    Get the the referred attribute.
    '''
    o_attr_ref = nav_one(o_attr).O_RATTR[106].O_BATTR[113].O_ATTR[106]()
    if o_attr_ref:
        return get_refered_attribute(o_attr_ref)
    else:
        return o_attr


def build_core_type(s_cdt):
    '''
    Build an xsd simpleType out of a S_CDT.
    '''
    s_dt = nav_one(s_cdt).S_DT[17]()
    
    if s_dt.name == 'void':
        type_name = None
    
    elif s_dt.name == 'boolean':
        type_name = 'xs:boolean'
    
    elif s_dt.name == 'integer':
        type_name = 'xs:integer'
    
    elif s_dt.name == 'real':
        type_name = 'xs:decimal'
    
    elif s_dt.name == 'string':
        type_name = 'xs:string'
    
    elif s_dt.name == 'unique_id':
        type_name = 'xs:integer'
    
    else:
        type_name = None
    
    if type_name:
        mapped_type = ET.Element('xs:simpleType', name=s_dt.name)
        ET.SubElement(mapped_type, 'xs:restriction', base=type_name)
        return mapped_type


def build_enum_type(s_edt):
    '''
    Build an xsd simpleType out of a S_EDT.
    '''
    s_dt = nav_one(s_edt).S_DT[17]()
    enum = ET.Element('xs:simpleType', name=s_dt.name)
    enum_list = ET.SubElement(enum, 'xs:restriction', base='xs:string')
    
    first_filter = lambda selected: not nav_one(selected).S_ENUM[56, 'succeeds']()
    
    s_enum = nav_any(s_edt).S_ENUM[27](first_filter)
    while s_enum:
        ET.SubElement(enum_list, 'xs:enumeration', value=s_enum.name)
        s_enum = nav_one(s_enum).S_ENUM[56, 'precedes']()
    
    return enum


def build_struct_type(s_sdt):
    '''
    Build an xsd complexType out of a S_SDT.
    '''
    s_dt = nav_one(s_sdt).S_DT[17]()
    struct = ET.Element('xs:complexType', name=s_dt.name)
    
    first_filter = lambda selected: not nav_one(selected).S_MBR[46, 'succeeds']()
    
    s_mbr = nav_any(s_sdt).S_MBR[44](first_filter)
    while s_mbr:
        s_dt = nav_one(s_mbr).S_DT[45]()
        type_name = get_type_name(s_dt)
        ET.SubElement(struct, 'xs:attribute', name=s_mbr.name, type=type_name)
        s_mbr = nav_one(s_mbr).S_MBR[46, 'precedes']()
    
    return struct


def build_user_type(s_udt):
    '''
    Build an xsd simpleType out of a S_UDT.
    '''
    s_dt_user = nav_one(s_udt).S_DT[17]()
    s_dt_base = nav_one(s_udt).S_DT[18]()
    
    base_name = get_type_name(s_dt_base)
    if base_name:
        user = ET.Element('xs:simpleType', name=s_dt_user.name)
        ET.SubElement(user, 'xs:restriction', base=base_name)
        
        return user


def build_type(s_dt):
    '''
    Build a partial xsd tree out of a S_DT and its sub types S_CDT, S_EDT, S_SDT and S_UDT.
    '''
    s_cdt = nav_one(s_dt).S_CDT[17]()
    if s_cdt:
        return build_core_type(s_cdt)
    
    s_edt = nav_one(s_dt).S_EDT[17]()
    if s_edt:
        return build_enum_type(s_edt)
    
    s_udt = nav_one(s_dt).S_UDT[17]()
    if s_udt:
        return build_user_type(s_udt)
    
#    s_sdt = nav_one(s_dt).S_SDT[17]()
#    if s_sdt:
#        return build_struct_type(s_sdt)


def build_class(o_obj):
    '''
    Build an xsd complex element out of a O_OBJ, including its O_ATTR.
    '''
    cls = ET.Element('xs:element', name=o_obj.key_lett, minOccurs='0', maxOccurs='unbounded')
    attributes = ET.SubElement(cls, 'xs:complexType')
    for o_attr in nav_many(o_obj).O_ATTR[102]():
        o_attr_ref = get_refered_attribute(o_attr)
        s_dt = nav_one(o_attr_ref).S_DT[114]()
        while nav_one(s_dt).S_UDT[17]():
            s_dt = nav_one(s_dt).S_UDT[17].S_DT[18]()
        
        type_name = get_type_name(s_dt)
        if type_name and not nav_one(o_attr).O_BATTR[106].O_DBATTR[107]():
            ET.SubElement(attributes, 'xs:attribute', name=o_attr.name, type=type_name)
        else:
            logger.warning('Omitting %s.%s' % (o_obj.key_lett, o_attr.Name))
    return cls


def build_component(m, c_c):
    '''
    Build an xsd complex element out of a C_C, including its packaged S_DT and O_OBJ.
    '''
    component = ET.Element('xs:element', name=c_c.name)
    
    classes = ET.SubElement(component, 'xs:complexType')
    classes = ET.SubElement(classes, 'xs:sequence')
    
    scope_filter = lambda selected: ooaofooa.is_contained_in(selected, c_c)
    
    for o_obj in m.select_many('O_OBJ', scope_filter):
        cls = build_class(o_obj)
        classes.append(cls)
    
    return component


def build_schema(m, c_c):
    '''
    Build an xsd schema from a bridgepoint component.
    '''
    schema = ET.Element('xs:schema')
    schema.set('xmlns:xs', 'http://www.w3.org/2001/XMLSchema')

    global_filter = lambda selected: ooaofooa.is_global(selected)
    for s_dt in m.select_many('S_DT', global_filter):
        datatype = build_type(s_dt)
        if datatype is not None:
            schema.append(datatype)
    
    scope_filter = lambda selected: ooaofooa.is_contained_in(selected, c_c)
    for s_dt in m.select_many('S_DT', scope_filter):
        datatype = build_type(s_dt)
        if datatype is not None:
            schema.append(datatype)
            
    component = build_component(m, c_c)
    schema.append(component)
    
    return schema


def prettify(xml_string):
    '''
    Indent an xml string with four spaces, and add an additional line break after each node.
    '''
    reparsed = xml.dom.minidom.parseString(xml_string)
    return reparsed.toprettyxml(indent="    ")


def main(args):
    
    parser = optparse.OptionParser(usage="%prog [options] <model_path> [another_model_path...]",
                                   version=xtuml.version.complete_string,
                                   formatter=optparse.TitledHelpFormatter())
    
    parser.set_description(__doc__.strip())
    
    parser.add_option("-c", "--component", dest="component", metavar="NAME",
                      help="export xsd schema for the component named NAME",
                      action="store", default=None)
    
    parser.add_option("-o", "--output", dest='output', metavar="PATH",
                      action="store", help="save xsd schema to PATH (required)",
                      default=None)
    
    parser.add_option("-v", "--verbosity", dest='verbosity', action="count", 
                      help="increase debug logging level", default=2)
    
    (opts, args) = parser.parse_args(args)
    if len(args) == 0 or None in [opts.component, opts.output]:
        parser.print_help()
        sys.exit(1)
        
    levels = {
              0: logging.ERROR,
              1: logging.WARNING,
              2: logging.INFO,
              3: logging.DEBUG,
    }
    logging.basicConfig(level=levels.get(opts.verbosity, logging.DEBUG))
    
    m = ooaofooa.load_metamodel(args)
    c_c = m.select_any('C_C', lambda inst: inst.Name == opts.component)
    if c_c:
        schema = build_schema(m, c_c)
        with open(opts.output, 'w') as f:
            s = ET.tostring(schema, 'utf-8')
            s = prettify(s)
            f.write(s)
    else:
        logger.error('unable to find a component named %s' % opts.component)
        logger.info('available components to choose from are: %s' %
                    ', '.join([c_c.Name for c_c in m.select_many('C_C')]))
        sys.exit(1)


if __name__ == '__main__':
    main(sys.argv[1:])
    
